"""Ledgers that record transactions."""
from .staking_ledger import StakingLedger
from .external_wallet_ledger import ExternalProtectionWalletLedger
from .pool_token_supply_ledger import PoolTokenSupplyLedger
from .vault_ledger import VaultLedger
from .available_liquidity_ledger import AvailableLiquidityLedger
from .vortex_ledger import VortexLedger



import sys
import pytest


def pytest_configure(config):
    config.addinivalue_line(
        "markers",
        "skip_offline: mark test to be skipped if offline (https://api.featurelabs.com cannot be reached)",
    )
    config.addinivalue_line(
        "markers", "noncore_dependency: mark test as needing non-core dependencies"
    )
    config.addinivalue_line(
        "markers",
        "skip_during_conda: mark test to be skipped if running during conda build",
    )
    config.addinivalue_line(
        "markers",
        "skip_if_39: mark test to be skipped if running during conda build",
    )


@pytest.fixture(scope="session")
def go():
    from plotly import graph_objects as go

    return go


@pytest.fixture(scope="session")
def im():
    from imblearn import over_sampling as im

    return im


@pytest.fixture(scope="session")
def graphviz():
    import graphviz

    return graphviz


def pytest_addoption(parser):
    parser.addoption(
        "--has-minimal-dependencies",
        action="store_true",
        default=False,
        help="If true, tests will assume only the dependencies in"
        "core-requirements.txt have been installed.",
    )
    parser.addoption(
        "--is-using-conda",
        action="store_true",
        default=False,
        help="If true, tests will assume that they are being run as part of"
        "the build_conda_pkg workflow with the feedstock.",
    )


def pytest_collection_modifyitems(config, items):
    if config.getoption("--has-minimal-dependencies"):
        skip_noncore = pytest.mark.skip(reason="needs noncore dependency")
        for item in items:
            if "noncore_dependency" in item.keywords:
                item.add_marker(skip_noncore)
    if config.getoption("--is-using-conda"):
        skip_conda = pytest.mark.skip(reason="Test does not run during conda")
        for item in items:
            if "skip_during_conda" in item.keywords:
                item.add_marker(skip_conda)
    if sys.version_info >= (3, 9):
        skip_39 = pytest.mark.skip(reason="Test dependency not supported in python 3.9")
        for item in items:
            if "skip_if_39" in item.keywords:
                item.add_marker(skip_39)


@pytest.fixture
def has_minimal_dependencies(pytestconfig):
    return pytestconfig.getoption("--has-minimal-dependencies")


@pytest.fixture
def is_using_conda(pytestconfig):
    return pytestconfig.getoption("--is-using-conda")


@pytest.fixture
def is_using_windows(pytestconfig):
    return sys.platform in ["win32", "cygwin"]


@pytest.fixture
def is_running_py_39_or_above():
    return sys.version_info >= (3, 9)


@pytest.fixture
def protocol():
    from bancorml.environments import Bancor3
    protocol = Bancor3
    return protocol





from .agent_base import AgentBase
# from .random_walk import RandomWalker
import random
import pandas as pd


class TraderAgent(AgentBase):
    """Trader subclass of RandomWalker, which is subclass to Mesa Agent
    """
    name = 'Trader Agent'

    def __init__(self,
                 env,
                 model=None,
                 pos=None,
                 moore=None,
                 unique_id=name
        ):

        # Bancor3 protocol, set at __init__, all LP's have the same protocol in this model
        self.protocol = env

        self.wallet = {
            'block_num': [0]
        }

        if model is not None:
            # init parent class with required parameters
            super().__init__(unique_id, pos, model, moore=moore)


    def swap(self, tkn1, x, tkn2, fake_wallet=True, block_num=0, update_liquidity=False):
        '''
            Loops over the addresses and discovers a random user with a non-zero balance of a random TKN (TKN).
            (conditional logic removes the blockNo and bnTKN lists from the choice)
            Creates a random number from the TKN balance (x).
            Adds new rows for each address, subtracts x from the appropriate list (i.e. as though the user has transferred it to the vault).
            Calls the AddLiquidity function, and adds the appropriate number of pool tokens to the appropriate list in the addresses dictionary.
            '''

        if tkn1 not in self.wallet and fake_wallet:
            self.wallet[tkn1]=[x * 1000 for i in range(len(self.wallet['block_num']))]
            self.wallet[tkn2]=[0 for i in range(len(self.wallet['block_num']))]

        elif tkn1 not in self.wallet:
            raise UserWarning(f"{tkn1} not found in agent wallet.")

        for key in self.wallet:
            if key == 'block_num':
                self.wallet[key].append(block_num)
            else:
                self.wallet[key].append(self.wallet[key][-1])

            if tkn1 == key:
                self.wallet[tkn1][-1] -= x
                self.wallet[tkn2][-1] += self.protocol.trade(tkn1, x, tkn2, block_num, update_liquidity)

            if (key != 'block_num') & (self.wallet[key][-1] < 0):
                if not self.protocol.warning_only:
                    raise UserWarning(f'Error. {key} balance cannot be less than zero. {self.wallet[key][-1]}')

    def get_wallet(self):
        return pd.DataFrame.from_dict(self.wallet)

    def step(self):
        # move to a cell in my Moore neighborhood
        self.random_move()



from .agent_base import AgentBase
# from .random_walk import RandomWalker
import random
import pandas as pd


class LiquidityProviderAgent(AgentBase):
    """ subclass of RandomWalker, which is subclass to Mesa Agent
    """
    name = 'Liquidity Provider Agent'

    def __init__(self,
                 env,
                 model=None,
                 pos=None,
                 moore=None,
                 unique_id=name,
        ):

        # Bancor3 protocol, set at __init__, all LP's have the same protocol in this model
        self.protocol = env
        self.wallet = {
            'block_num': [0]
        }

        if model is not None:
            # init parent class with required parameters
            super().__init__(unique_id, pos, model, moore=moore)


    def deposit(self, tkn, amount, block_num=0, fake_wallet=True):

        if fake_wallet:
            self.handle_fake_wallet(tkn, amount, m=1000)
        else:
            self.wallet[f'bn{tkn}'] = [0]

        for key in self.wallet:
            if (tkn == key) & (tkn == 'BNT'):
                self.wallet[tkn].append(self.wallet[tkn][-1] - amount)
                self.wallet[f"v{tkn}"].append(self.wallet[f"v{tkn}"][-1] + self.protocol.stake(tkn, amount, block_num))
                self.wallet[f"bn{tkn}"].append(self.wallet[f"bn{tkn}"][-1] + self.protocol.stake(tkn, amount, block_num))
            elif tkn == key:
                self.wallet[tkn].append(self.wallet[tkn][-1] - amount)
                self.wallet[f"bn{tkn}"].append(self.wallet[f"bn{tkn}"][-1] + self.protocol.stake(tkn, amount, block_num))
            elif (key.replace('bn','') not in self.wallet) & (key != 'block_num'):
                self.wallet[key].append(self.wallet[key][-1])
                self.wallet[f"bn{key}"].append(self.wallet[f"bn{key}"][-1])
            elif key == 'block_num':
                self.wallet[key].append(self.wallet[key][-1])

            if (key != 'block_num') & (self.wallet[key][-1] < 0):
                if not self.protocol.warning_only:
                    raise UserWarning(f'Error. {key} balance cannot be less than zero. {self.wallet[key][-1]}')


        if len(self.wallet['BNT']) < len(self.wallet[tkn]):
            self.wallet['BNT'].append(self.wallet['BNT'][-1])

    def withdraw(self, tkn, amount, block_num=0, fake_wallet=True):
        '''
        Loops over the addresses and discovers a random user with a non-zero balance of a random TKN (TKN).
        (conditional logic removes the blockNo and bnTKN lists from the choice)
        Creates a random number from the TKN balance (x).
        Adds new rows for each address, subtracts x from the appropriate list (i.e. as though the user has transferred it to the vault).
        Calls the AddLiquidity function, and adds the appropriate number of pool tokens to the appropriate list in the addresses dictionary.
        '''
        if fake_wallet:
            self.handle_fake_wallet(tkn, amount, m=0)
        else:
            self.wallet[f'bn{tkn}'] = [0]

        if 'BNT' not in self.wallet:
            self.wallet['BNT'] = [0 for i in range(len(self.wallet[tkn]))]

        for key in self.wallet:
            if (tkn == key) & (tkn == 'BNT'):
                tkn_out, bnt_out = self.protocol.unstake(tkn, amount)
                self.wallet["BNT"].append(self.wallet["BNT"][-1] + tkn_out)
                self.wallet["BNT"].append(self.wallet["BNT"][-1] + bnt_out)
            elif tkn == key:
                tkn_out, bnt_out = self.protocol.unstake(tkn, amount)
                self.wallet[tkn].append(self.wallet[tkn][-1] + tkn_out)
                self.wallet["BNT"].append(self.wallet["BNT"][-1] + bnt_out)
            else:
                if len(self.wallet[key]) < len(self.wallet[tkn]):
                    self.wallet[key].append(self.wallet[key][-1])

            if (key != 'block_num') & (self.wallet[key][-1] < 0):
                if not self.protocol.warning_only:
                    raise UserWarning(f'Error. {key} balance cannot be less than zero. {self.wallet[key][-1]}')


        if len(self.wallet['block_num']) < len(self.wallet[tkn]):
            self.wallet['block_num'].append(block_num)
        if len(self.wallet['BNT']) < len(self.wallet[tkn]):
            self.wallet['BNT'].append(self.wallet['BNT'][-1])

    def handle_fake_wallet(self, tkn, amount, m):
        if f'bn{tkn}' not in self.wallet:
            self.wallet[tkn]=[amount * m for i in range(len(self.wallet['block_num']))]
            self.wallet[f'bn{tkn}']=[0 for i in range(len(self.wallet['block_num']))]
            if tkn == 'BNT':
                self.wallet[f'v{tkn}'] = [0 for i in range(len(self.wallet['block_num']))]

        if 'BNT' not in self.wallet:
            self.wallet['BNT'] = [0 for i in range(len(self.wallet['block_num']))]

    def get_wallet(self, tkn=None, history=False):
        if (tkn is None) & (not history):
            return pd.DataFrame.from_dict(self.wallet).iloc[[-1]]
        elif (tkn is None) and history:
            return pd.DataFrame.from_dict(self.wallet)
        else:
            bntkn_val = pd.DataFrame.from_dict(self.wallet).iloc[[-1]][f'bn{tkn}'].values[0]
            tkn_val = pd.DataFrame.from_dict(self.wallet).iloc[[-1]][tkn].values[0]
            return f'{tkn}={tkn_val}', f'bn{tkn}={bntkn_val}'

    def step(self):
        # move to a cell in my Moore neighborhood
        self.random_move()



"""Utility methods."""
from .logger import get_logger, log_subtitle, log_title
from .crypto_utils import (
    iloss,
    compare,
    get_gecko_spot_price,
    gecko_list,
    geckoHistorical
)
from .cli_utils import (
    get_bancorml_root,
    get_installed_packages,
    get_sys_info,
    print_deps,
    print_info,
    print_sys_info,
)
from .data_utils import (
    decimal_from_value,
    load_data,
    converters,
    parse_json_tests,
    load_test_data
)
from .schemas import (
    FixedPointUnstakeTKN,
    FloatingPointUnstakeTKN
)


"""bancorml."""
import warnings
import bancorml.environments
import bancorml.utils

with warnings.catch_warnings():
    warnings.simplefilter("ignore", FutureWarning)
    warnings.simplefilter("ignore", DeprecationWarning)

warnings.filterwarnings("ignore", category=FutureWarning)
warnings.filterwarnings("ignore", category=DeprecationWarning)

__version__ = "0.2.16"



from bancorml.environments.env_base import EnvBase
from bancorml.agents import LiquidityProviderAgent


class Bancor3(EnvBase):
    """Bancor3 Environment.

        Args:
            is_solidity (bool): Toggles between floating-point-signed and fixed-point-unsigned
            min_liquidity_threshold (float): The BancorDAO prescribes a liquidity threshold, denominated in BNT, that represents the minimum available liquidity that must be present before the protocol bootstraps the pool with BNT.
            bnt_funding_limit (int): The BancorDAO determines the available liquidity for trading, through adjustment of the “BNT funding limit” parameter.
            alpha (float): Alpha value in the EMA equation.
            pool_fees (dict): Dictionary of tkn:fee (str:float) key-values per pool.
            cooldown_period (int): The cooldown period in days.
            exit_fee (float): The global exit fee.
            spot_prices (dict): Dictionary of tkn:price (str:float) key-values per tkn.
            whitelisted_tokens (list): List of token tickernames indicating whitelist status approval.
            bootstrapped_tokens (list): List of DAOmsig initiated pools/tokens.
            vortex_rates (dict): Dictionary of tkn:rate (str:float) key-values per TKN.
        """

    name = "Bancor3 Environment"
    block_num = 0
    ema = {
        'TKN': {'block_num': [0], 'ema': [0]},
        'BNT': {'block_num': [0], 'ema': [0]},
        'DAI': {'block_num': [0], 'ema': [0]},
        'wBTC': {'block_num': [0], 'ema': [0]},
        'ETH': {'block_num': [0], 'ema': [0]},
        'LINK': {'block_num': [0], 'ema': [0]}
    }

    def __init__(
            self,
            block_num=0,
            alpha=0.2,
            is_solidity=False,
            min_liquidity_threshold=10000,
            exit_fee=.0025,
            cooldown_period=7,
            bnt_funding_limit=100000,
            external_price_feed={
                        "TKN": {'block_num': [0, 1, 2, 3],
                                'price_usd': [2.50, 2.49, 2.51, 2.50]},
                        "BNT": {'block_num': [0, 1, 2, 3],
                                'price_usd': [2.50, 2.49, 2.51, 2.50]},
                        "LINK": {'block_num': [0, 1, 2, 3],
                                'price_usd': [15.35, 15.20, 15.11, 15.00]},
                        "ETH": {'block_num': [0, 1, 2, 3],
                                'price_usd': [2550.00, 2400.00, 2450.00, 2500.00]},
                        "wBTC": {'block_num': [0, 1, 2, 3],
                                'price_usd': [40123.23, 40312.21, 40111.11, 40000.00]},
                        "DAI": {'block_num': [0, 1, 2, 3],
                                'price_usd': [1.0, 1.0, 1.0, 1.00]}
            },
            pool_fees={
                "TKN": 0.01,
                "BNT": 0.01,
                "LINK": 0.01,
                "ETH": 0.01,
                "wBTC": 0.01,
                "DAI": 0.01,
            },
            whitelisted_tokens=['wBTC','BNT','LINK','ETH','DAI','TKN'],
            bootstrapped_tokens=[],
            exchange_rates={},
            spot_prices={
                "TKN": {'block_num': [0],
                        'price_usd': [2.50]},
                "BNT": {'block_num':[0],
                        'price_usd':[2.50]},
                "LINK": {'block_num':[0],
                        'price_usd':[15.00]},
                "ETH": {'block_num':[0],
                        'price_usd':[2500.00]},
                "wBTC": {'block_num':[0],
                        'price_usd':[40000.00]},
                "DAI": {'block_num': [0],
                         'price_usd': [1.00]}
            },
            vortex_rates={
                "TKN": 0.2,
                "BNT": 0.2,
                "LINK": 0.2,
                "ETH": 0.2,
                "wBTC": 0.2,
                "DAI": 0.2,
            },
            warning_only=False,
            dao_msig_initialized_pools=[],
            verbose=True,
            **kwargs
    ):

        super().__init__()

        self.set_param(alpha=alpha,
                       warning_only=warning_only,
                       spot_prices=spot_prices,
                       vortex_rates=vortex_rates,
                       block_num=block_num,
                       exchange_rates=exchange_rates,
                       min_liquidity_threshold=min_liquidity_threshold,
                       bnt_funding_limit=bnt_funding_limit,
                       exit_fee=exit_fee,
                       cooldown_period=cooldown_period,
                       pool_fees=pool_fees,
                       bootstrapped_tokens=bootstrapped_tokens,
                       whitelisted_tokens=whitelisted_tokens,
                       external_price_feed=external_price_feed,
                       is_solidity=is_solidity,
                       dao_msig_initialized_pools=dao_msig_initialized_pools,
                       verbose=verbose)

        for tkn in self.available_liquidity_ledger:
            self.available_liquidity_ledger[tkn]['funding_limit'][-1] = bnt_funding_limit

        self.protocol_agent = LiquidityProviderAgent(env=self, unique_id="Bancor Protocol")
        self._ema = Bancor3.ema

    def unstake(self, tkn, x, p=0, q=0, r=0, s=0, t=0, u=0):
        """Bancor3 Unstake protocol logic

        Args:
            x (int or float): The precise TKN value of the bnTKN tokens being withdrawn.
            uid (str or int): The agent user id (default=0)
            tkn (str): The abbreviated token name (default=BNT, valid options are as listed on CoinGecko)

        Returns:
            tkn_out (int or float or decimal): The x of TKN to return to the user
            bnt_out (int or float or decimal): The x of BNT to return to the user
        """
        if tkn == 'BNT':
            exchange_rate =  self.staking_ledger['BNT'][-1] / self.pool_token_supply_ledger['bnBNT_ERC20_contract']['supply'][-1]
            bnt = exchange_rate * x
            exit_fee = self.exit_fee * bnt
            tkn_out = 0
            bnt_out = bnt - exit_fee
        else:
            a, b, c, e, m, n, x, w = self.validate_input_types(tkn, x)

            # check balanced
            self.tokenomics.check_balanced(c)

            if self.tokenomics.is_balanced:
                case = 'bootstrap surplus'

            else:

                # check surplus
                self.tokenomics.check_surplus(b, c, e, n)

                # check hlim
                self.tokenomics.check_hlim(b, c, e, x)

                # check hmax
                self.tokenomics.check_hmax(b, c, e, m, n, x, self.tokenomics.surplus)

                # check if trading liquidity needs reduced
                self.tokenomics.check_reduce_trading_liquidity(x, b, c, e, n,
                                                               self.tokenomics.surplus,
                                                               self.tokenomics.satisfies_hmax,
                                                               self.tokenomics.satisfies_hlim)

                # check case (e.g. "default surplus", "arbitrage deficit", etc...)
                case = self.check_case(self.tokenomics.surplus,
                                       self.tokenomics.deficit,
                                       self.tokenomics.satisfies_hlim,
                                       self.tokenomics.satisfies_hmax,
                                       self.tokenomics.reduce_trading_liquidity,
                                       a, b, c, e, w)

            if case in ['bootstrap surplus', 'bootstrap surplus (special case)']:
                s = self.tokenomics.handle_bootstrap_surplus(x, n)

            elif case == 'bootstrap deficit (special case)':
                s = self.tokenomics.handle_bootstrap_deficit_special_case(c, e, n, x)

            elif case == 'arbitrage surplus':
                p, r, s = self.tokenomics.handle_arbitrage_surplus(x, n, a, b, c, e, m)

            elif case == 'default surplus':
                p, q, r, s = self.tokenomics.handle_default_surplus(x, n, c, a, b)

            elif case == 'arbitrage deficit':
                p, r, s = self.tokenomics.handle_arbitrage_deficit(x, n, a, m, e, b, c)

            elif case == 'bootstrap deficit':
                s, t = self.tokenomics.handle_bootstrap_deficit(x, n, b, c, e, a)

            elif case == 'default deficit':
                p, q, r, s, t = self.tokenomics.handle_default_deficit(a, b, c, e, n, x)

            if tkn in self.external_protection_wallet_ledger:
                external_wallet_case = self.tokenomics.check_external_wallet_adjustment(a, b, w, t, self.tokenomics.satisfies_hlim, self.tokenomics.satisfies_hmax)

                if external_wallet_case == 'external wallet adjustment (1)':
                    t, u = self.tokenomics.handle_external_wallet_adjustment_1(t, b, a)
                    case = case + ' ep1'

                elif external_wallet_case == 'external wallet adjustment (2)':
                    t, u = self.tokenomics.handle_external_wallet_adjustment_2(w, t, b, a)
                    case = case + ' ep2'

            # update ledger balances based on the current state
            self.update_ledger_balances(tkn, p, q, r, s, t, u, a, b, c, e, m, n, x, w, case)
            tkn_out = s + u
            bnt_out = t

        self.check_pool_shutdown(warning_only=self.warning_only)
        return tkn_out, bnt_out

    def stake(self, tkn, x, block_num):
        """Main protocol logic to perform staking functionality

        Args:
            tkn (str): The abbreviated token name (default=BNT, valid options are as listed on CoinGecko)
            x (int or float): The precise TKN value of the bnTKN tokens being withdrawn.

        Returns:
            y (int or float or decimal): the number of pool tokens issued to the user.
        """
        if block_num != self.block_num:
            self.update_ema(tkn)
            self.block_num = block_num

        assert tkn in self.whitelisted_tokens, 'The token is not supported'
        if tkn == 'BNT':
            exchange_rate =   self.pool_token_supply_ledger['bnBNT_ERC20_contract']['supply'][-1] / self.staking_ledger['BNT'][-1]
            y = exchange_rate * x
        else:
            y = self.mint_bntkn(tkn, x)
            self.add_tkn_to_vault(tkn, x)
            self.add_tkn_to_staking_ledger(tkn, x)
            if self.is_within_ema_tolerance(tkn) and (tkn in self.dao_msig_initialized_pools):
                self.add_tkn_to_available_liquidity_ledger(tkn, x)
            self.check_bootstrapped_tokens()

        self.check_pool_shutdown(warning_only=self.warning_only)
        return (y)

    def trade(self, tkn1, x, tkn2, block_num, update_liquidity=False):
        """Main protocol logic to perform trading functionality.

        Args:
            tkn1 (str): The abbreviated token name being traded-in
            x (int or float): The precise TKN value of tkn1 being traded.
            tkn2 (str): The abbreviated token name being traded-for (i.e., sent back to the user)

        Returns:
            tkn_out (int or float or decimal): the number of pool tokens issued to the user.
        """
        if block_num != self.block_num:
            self.update_ema(tkn1)
            self.update_ema(tkn2)
            self.block_num = block_num

        if (tkn1=='BNT') or (tkn2=='BNT'):

            if tkn1=='BNT':
                tkn_a = tkn2
                operation = self.tokenomics.handle_trade_bnt_to_tkn

            elif tkn2=='BNT':
                tkn_a = tkn1
                operation = self.tokenomics.handle_trade_tkn_to_bnt

            a = self.available_liquidity_ledger[tkn_a]['BNT'][-1]
            b = self.available_liquidity_ledger[tkn_a][tkn_a][-1]
            d = self.pool_fees[tkn_a]
            e = self.vortex_rates[tkn_a]
            bnt_trading_liquidity, tkn_trading_liquidity, tkn_out, tkn_fee, vortex_fee = operation(
                a, d, b, x, e)

            if self.is_within_ema_tolerance(tkn_a) or update_liquidity:
                self.update_available_liquidity_ledger(tkn_a, bnt_trading_liquidity, tkn_trading_liquidity)

            self.add_tkn_to_vault(tkn1, x)
            self.add_tkn_to_vault(tkn2, -tkn_out)
            self.add_tkn_to_staking_ledger(tkn2, tkn_fee)
            self.add_tkn_to_vortex_ledger(vortex_fee)

            print(f'bnt_trading_liquidity={bnt_trading_liquidity}, \n'
                  f'tkn_trading_liquidity={tkn_trading_liquidity}, \n'
                  f'tkn_out={tkn_out}, \n'
                  f'tkn_fee={tkn_fee}, \n'
                  f'vortex_fee={vortex_fee} \n\n')

        else:

            a1 = self.available_liquidity_ledger[tkn1]['BNT'][-1]
            b1 = self.available_liquidity_ledger[tkn1][tkn1][-1]
            d1 = self.pool_fees[tkn1]

            a2 = self.available_liquidity_ledger[tkn2]['BNT'][-1]
            b2 = self.available_liquidity_ledger[tkn2][tkn2][-1]
            d2 = self.pool_fees[tkn2]

            e = self.vortex_rates[tkn1]

            bnt_source_trading_liquidity, tkn_source_trading_liquidity, bnt_destination_trading_liquidity, tkn_destination_trading_liquidity, tkn_out, bnt_fee, tkn_fee, vortex_fee = self.tokenomics.handle_trade_tkn_to_tkn(
                a1, d1, b1, a2, d2, b2, x, e)

            if self.is_within_ema_tolerance(tkn1):
                self.update_available_liquidity_ledger(tkn1, bnt_source_trading_liquidity, tkn_source_trading_liquidity)

            if self.is_within_ema_tolerance(tkn2):
                self.update_available_liquidity_ledger(tkn2, bnt_destination_trading_liquidity, tkn_destination_trading_liquidity)

            self.add_tkn_to_vault(tkn1, x)
            self.add_tkn_to_vault(tkn2, -tkn_out)
            self.add_tkn_to_staking_ledger(tkn2, tkn_fee)
            self.add_tkn_to_staking_ledger('BNT', bnt_fee)
            self.add_tkn_to_vortex_ledger(vortex_fee)

            print(f'bnt_source_trading_liquidity={bnt_source_trading_liquidity}, \n',
                  f'tkn_source_trading_liquidity={tkn_source_trading_liquidity}, \n',
                  f'bnt_destination_trading_liquidity={bnt_destination_trading_liquidity}, \n',
                  f'tkn_destination_trading_liquidity={tkn_destination_trading_liquidity}, \n',
                  f'tkn_out={tkn_out}, \n',
                  f'bnt_fee={bnt_fee}, \n',
                  f'tkn_fee={tkn_fee}, \n',
                  f'vortex_fee={vortex_fee} \n')

        self.check_pool_shutdown(warning_only=self.warning_only)

        for token in [tkn1, tkn2]:
            self.update_spot_prices(token)

        return tkn_out




# import numpy as np
#
# from gym import Space
#
# class ObservationSpaceDAO(Space):
#     max_ratio = 3.0
#
#     def __init__(self):
#         super(ObservationSpaceDAO, self).__init__()
#
#     def sample(self):
#         return np.random.uniform(0, ObservationSpaceDAO.max_ratio, 4)
#
#     def contains(self, obs, x):
#         return len(obs) == 4 and (obs >= 0.0).all() and (x <= ObservationSpaceDAO.max_ratio).all()
#
#     def to_jsonable(self, sample_n):
#         return np.array(sample_n).to_list()
#
#     def from_jsonable(self, sample_n):
#         return [np.asarray(sample) for sample in sample_n]
#
#
# class ObservationSpaceTrader(Space):
#     max_ratio = 3.0
#
#     def __init__(self):
#         super(ObservationSpaceTrader, self).__init__()
#
#     def sample(self):
#         return np.random.uniform(0, ObservationSpaceTrader.max_ratio, 4)
#
#     def contains(self, obs, x):
#         return len(obs) == 4 and (obs >= 0.0).all() and (x <= ObservationSpaceTrader.max_ratio).all()
#
#     def to_jsonable(self, sample_n):
#         return np.array(sample_n).to_list()
#
#     def from_jsonable(self, sample_n):
#         return [np.asarray(sample) for sample in sample_n]
#
#
# class ObservationSpaceArbitrageur(Space):
#     max_ratio = 3.0
#
#     def __init__(self):
#         super(ObservationSpaceArbitrageur, self).__init__()
#
#     def sample(self):
#         return np.random.uniform(0, ObservationSpaceArbitrageur.max_ratio, 4)
#
#     def contains(self, obs, x):
#         return len(obs) == 4 and (obs >= 0.0).all() and (x <= ObservationSpaceArbitrageur.max_ratio).all()
#
#     def to_jsonable(self, sample_n):
#         return np.array(sample_n).to_list()
#
#     def from_jsonable(self, sample_n):
#         return [np.asarray(sample) for sample in sample_n]
#
#
# observation_space = {
#     'DAO': ObservationSpaceDAO,
#     'Trader': ObservationSpaceTrader,
#     'Arbitrageur': ObservationSpaceArbitrageur
# }


"""Base Agent Environment Class."""
import logging
from abc import ABC, abstractmethod
from decimal import Decimal
import numpy as np
import pandas as pd
# from bancorml.environments.observation_space import observation_space
# from evalml.utils.logger import get_logger, log_title, log_subtitle
from bancorml.environments.tokenomics.ledgers import VaultLedger, PoolTokenSupplyLedger, ExternalProtectionWalletLedger, \
    StakingLedger, AvailableLiquidityLedger, VortexLedger
from bancorml.utils import FixedPointUnstakeTKN, FloatingPointUnstakeTKN
from bancorml.environments.tokenomics.floating_point_tokenomics import FloatingPointTokenomics
from bancorml.environments.tokenomics.fixed_point_tokenomics import FixedPointTokenomics

# logger = logging.getLogger(__name__)

class EnvBase(ABC):
    """Environment base for multi-agent-oriented system.

    Args:
        observation_space (obj): Gym.Spaces class instance which defines respective agent-oriented observation spaces
        random_seed (int): Seed for the random number generator. Defaults to 0.
    """
    metadata = {'render.modes': []}


    def __init__(
        self,
        # observation_space=observation_space,
    ):
        # self.observation_space = observation_space
        # self._external_price_feed = external_price_feed
        # self.logger = get_logger(f"{__name__}")

        # create a vault
        self.vault_ledger = VaultLedger().ledger

        # create a staking ledger
        self.staking_ledger = StakingLedger().ledger

        # create a single liquidity pool for the model
        self.pool_token_supply_ledger = PoolTokenSupplyLedger().ledger

        # create a external protection wallet for the model
        self.external_protection_wallet_ledger = ExternalProtectionWalletLedger().ledger

        # create a ledger to track available liquidity
        self.available_liquidity_ledger = AvailableLiquidityLedger().ledger

        self.vortex_ledger = VortexLedger().ledger
        self.actions = 0

    @property
    def min_liquidity_threshold(self) -> int or float:
        """Minimum pool liquidity threshold before the DAOmsig can initiate."""
        return self._min_liquidity_threshold

    @min_liquidity_threshold.setter
    def min_liquidity_threshold(self, updated_min_liquidity_threshold: int or float):
        self._min_liquidity_threshold = updated_min_liquidity_threshold

    @property
    def warning_only(self) -> bool:
        """If True then will cause exception and break whenever a negative balance is found."""
        return self._warning_only

    @warning_only.setter
    def warning_only(self, updated_warning_only: bool):
        self._warning_only = updated_warning_only

    @property
    def bnt_funding_limit(self) -> int or float:
        """The BancorDAO determines the available liquidity for trading, through adjustment of the “BNT funding limit” parameter."""
        return self._bnt_funding_limit

    @bnt_funding_limit.setter
    def bnt_funding_limit(self, updated_bnt_funding_limit: int or float):
        self._bnt_funding_limit = updated_bnt_funding_limit

    @property
    def alpha(self) -> float:
        """EMA calculation alpha setting"""
        return self._alpha

    @alpha.setter
    def alpha(self, updated_alpha: float):
        self._alpha = updated_alpha

    @property
    def verbose(self) -> bool:
        """Whether to print system state during processing"""
        return self._verbose

    @verbose.setter
    def verbose(self, updated_verbose: bool):
        self._verbose = updated_verbose

    @property
    def whitelisted_tokens(self) -> list:
        """List of approved tokens allowed on Bancor v3"""
        return self._whitelisted_tokens

    @whitelisted_tokens.setter
    def whitelisted_tokens(self, updated_whitelist: list):
        self._whitelisted_tokens = updated_whitelist

    @property
    def dao_msig_initialized_pools(self) -> list:
        """List of DAOmsig initialized pools on Bancor v3"""
        return self._dao_msig_initialized_pools

    @dao_msig_initialized_pools.setter
    def dao_msig_initialized_pools(self, updated_dao_msig_initialized_pools: list):
        self._dao_msig_initialized_pools = updated_dao_msig_initialized_pools

    @property
    def external_price_feed(self) -> dict:
        """List of approved tokens allowed on Bancor v3"""
        return self._external_price_feed

    @external_price_feed.setter
    def external_price_feed(self, updated_external_price_feed: dict):
        self._external_price_feed = updated_external_price_feed

    @property
    def bootstrapped_tokens(self) -> list:
        """List of whitelisted tkn pools on Bancor v3 which meet the minimum pool liquidity threshold"""
        return self._bootstrapped_tokens

    @bootstrapped_tokens.setter
    def bootstrapped_tokens(self, updated_bootstrapped_tokens: list):
        self._bootstrapped_tokens = updated_bootstrapped_tokens

    @property
    def vortex_rates(self) -> dict:
        """The Bancor Vortex accrues value from its share of the revenues is used to buy and burn vBNT to provide IL protection."""
        return self._vortex_rates

    @vortex_rates.setter
    def vortex_rates(self, updated_vortex_rates: dict):
        self._vortex_rates = updated_vortex_rates

    @property
    def pool_fees(self) -> dict:
        """Protocol revenue (i.e. the pool fees, or the commissions paid by traders) is distributed to users by increasing the balance of the staking ledger, equal to the value captured."""
        return self._pool_fees

    @pool_fees.setter
    def pool_fees(self, updated_pool_fees: dict):
        self._pool_fees = updated_pool_fees

    @property
    def exit_fee(self) -> Decimal:
        """
        As a final circumspection against bad behavior,
        exit fees are introduced into the Bancor ecosystem for the first time.
        The exit fee is designed to temper the profit motive of the imagined exploit vector.
        """
        return self._exit_fee

    @exit_fee.setter
    def exit_fee(self, updated_exit_fee: Decimal):
        self._exit_fee = updated_exit_fee

    @property
    def exchange_rates(self) -> dict:
        """BNT to TKN exchange rate, effectively swapping out an equal quantity of BNT value for TKN value, until there is no TKN remaining."""
        return self._exchange_rates

    @exchange_rates.setter
    def exchange_rates(self, updated_exchange_rates: dict):
        self._exchange_rates = updated_exchange_rates

    @property
    def ema(self) -> dict:
        """Exponential Moving Average (EMA) used to test spot price tolerance."""
        return self._ema

    @ema.setter
    def ema(self, updated_ema: dict):
        self._ema = updated_ema

    @property
    def spot_prices(self) -> dict:
        """Current system state spot prices. Setting these will cause an auto-update"""
        return self._spot_prices

    @spot_prices.setter
    def spot_prices(self, updated_spot_prices: dict):
        self._spot_prices = updated_spot_prices

    @property
    def custom_name(self):
        """Custom name of the env."""
        return self._custom_name

    @property
    def name(self):
        """Name of the env."""
        return self.custom_name or self.summary

    def set_param(self, is_solidity=None,
                        block_num=None,
                        min_liquidity_threshold=None,
                        bnt_funding_limit=None,
                        alpha=None,
                        pool_fees=None,
                        cooldown_period=None,
                        exit_fee=None,
                        spot_prices=None,
                        whitelisted_tokens=None,
                        bootstrapped_tokens=None,
                        exchange_rates=None,
                        external_price_feed=None,
                        vortex_rates=None,
                        dao_msig_initialized_pools=None,
                        warning_only=None,
                        verbose=None
                  ):
        """Allows user to set the specified system parameters.

        Args:
            is_solidity (bool): Toggles between floating-point-signed and fixed-point-unsigned
            min_liquidity_threshold (float): The BancorDAO prescribes a liquidity threshold, denominated in BNT, that represents the minimum available liquidity that must be present before the protocol bootstraps the pool with BNT.
            bnt_funding_limit (int): The BancorDAO determines the available liquidity for trading, through adjustment of the “BNT funding limit” parameter.
            alpha (float): Alpha value in the EMA equation.
            pool_fees (dict): Dictionary of tkn:fee (str:float) key-values per pool.
            cooldown_period (int): The cooldown period in days.
            exit_fee (float): The global exit fee.
            spot_prices (dict): Dictionary of tkn:price (str:float) key-values per tkn.
            whitelisted_tokens (list): List of token tickernames indicating whitelist status approval.
            bootstrapped_tokens (list): List of DAOmsig initiated pools/tokens.
            vortex_rates (dict): Dictionary of tkn:rate (str:float) key-values per TKN.

        Returns:
            self
        """
        if block_num is not None:
            self.block_num = block_num
        if exchange_rates is not None:
            self.exchange_rates = exchange_rates
        if min_liquidity_threshold is not None:
            self.min_liquidity_threshold = min_liquidity_threshold
        if bnt_funding_limit is not None:
            self.bnt_funding_limit = bnt_funding_limit
        if alpha is not None:
            self.alpha = alpha
        if warning_only is not None:
            self.warning_only = warning_only
        if cooldown_period is not None:
            self.cooldown_period = cooldown_period
        if pool_fees is not None:
            self.pool_fees = pool_fees
        if exit_fee is not None:
            self.exit_fee = exit_fee
        if whitelisted_tokens is not None:
            self.set_whitelisted_tokens(whitelisted_tokens)
        if vortex_rates is not None:
            self.vortex_rates = vortex_rates
        if external_price_feed is not None:
            self.external_price_feed = external_price_feed
        if bootstrapped_tokens is not None:
            self.bootstrapped_tokens = bootstrapped_tokens
        if dao_msig_initialized_pools is not None:
            self.dao_msig_initialized_pools = dao_msig_initialized_pools
        if verbose is not None:
            self.verbose = verbose
        if is_solidity is not None:
            self.is_solidity = is_solidity
            self.tokenomics = FixedPointTokenomics() if is_solidity else FloatingPointTokenomics()
        if spot_prices is not None:
            self.spot_prices = spot_prices

        self.update_exchange_rates()
        self.init_ema()

    def init_ema(self):
        ema = self.ema
        for tkn in self.spot_prices:
            if tkn in ema:
                ema[tkn]['block_num'] = [self.block_num]
                ema[tkn]['ema'] = [self.exchange_rates[tkn]]
        self.ema = ema

    def update_exchange_rates(self):
        exchange_rates = self.exchange_rates
        bnt_price = self.spot_prices['BNT']['price_usd'][-1]
        for tkn in self.spot_prices:
            tkn_price = self.spot_prices[tkn]['price_usd'][-1]
            exchange_rates[tkn] = tkn_price / bnt_price
        self.exchange_rates = exchange_rates

    def get_available_liquidity_ledger(self, tkn):
        """Gets the available trading liquidity ledger.

        Args:
            tkn (str): Tickername of the token of interest.
        Returns:
            (pd.DataFrame): DataFrame of available_liquidity_ledger logging.
        """
        return pd.DataFrame.from_dict(self.available_liquidity_ledger[tkn])

    def get_vortex_ledger(self):
        """Gets the vortex ledger.

        Returns:
            (pd.DataFrame): DataFrame of vortex logging.
        """
        return pd.DataFrame.from_dict(self.vortex_ledger)

    def get_vault_ledger(self):
        """Gets the vault ledger.

        Returns:
            (pd.DataFrame): DataFrame of vault ledger logging.
        """
        return pd.DataFrame.from_dict(self.vault_ledger)

    def get_staking_ledger(self):
        """Gets the staking ledger.

        Returns:
            (pd.DataFrame): DataFrame of staking ledger logging.
        """
        return pd.DataFrame.from_dict(self.staking_ledger)

    def get_pool_token_supply_ledger(self, tkn):
        """Gets the pool token supply ledger.

        Args:
            tkn (str): Tickername of the token of interest.
        Returns:
            (pd.DataFrame): DataFrame of pool token supply ledger logging.
        """
        return pd.DataFrame.from_dict(self.pool_token_supply_ledger[f'bn{tkn}_ERC20_contract'])

    def get_pool_token_value(self, tkn):
        """Get pool token value in bnTKN/TKN. If this is the first issuance, the pool token value is forced to 1.

        Args:
            x (int or float or decimal): The value of BNT to add to the vortex
        Returns:
            bnTKN_value (int or float or decimal): value of a single bnTKN pool token.
        """
        staked_amount = self.staking_ledger[tkn][-1]
        bnTKN_supply = self.pool_token_supply_ledger[f'bn{tkn}_ERC20_contract']['supply'][-1]
        if staked_amount > 0 and bnTKN_supply > 0:
            bnTKN_value = float(bnTKN_supply) / float(staked_amount)
        else:
            bnTKN_value = 1
        return bnTKN_value

    def update_spot_prices(self, tkn):
        """Updates the internal spot price as determined by the trading liquidity balances of the pool

        Args:
            tkn (str): The value of BNT to add to the vortex
        Returns:
            self
        """
        spot_rate = self.get_spot_rate(tkn)
        spot_prices = self.spot_prices
        spot_prices[tkn]['block_num'].append(self.block_num)
        spot_prices[tkn]['price_usd'].append(spot_rate)
        self.spot_prices = spot_prices
        self.update_exchange_rates()

    def get_spot_rate(self, tkn, update_spot_prices=False):
        """Gets the spot rate in units of BNT/TKN as determined by the trading liquidity balances of the pool

        Args:
            tkn (str): The value of BNT to add to the vortex
        Returns:
            spot_rate (int or float or decimal): spot rate in units of BNT/TKN
        """
        if tkn == 'BNT':
            spot_rate = self.exchange_rates[tkn]
        else:
            TKN_amt = self.available_liquidity_ledger[tkn][tkn][-1]
            BNT_amt = self.available_liquidity_ledger[tkn]['BNT'][-1]
            if TKN_amt > 0 and BNT_amt > 0:
                spot_rate = BNT_amt / TKN_amt
            else:
                spot_rate = self.exchange_rates[tkn]

        if update_spot_prices:
            spot_prices = self.spot_prices
            spot_prices[tkn]['block_num'].append(self.block_num)
            spot_prices[tkn]['price_usd'].append(spot_rate)
            self.spot_prices = spot_prices
            self.update_exchange_rates()
        return spot_rate

    def get_bnt_rate(self, tkn) -> float:
        """Gets the rate in units of TKN/BNT as determined by the external price feed

        Args:
            tkn (str): The token tickername.
        Returns:
            bnt_rate (int or float or decimal): bnt rate in units of TKN/BNT
        """
        bnt_price = self.external_price_feed['BNT']['price_usd'][-1]
        tkn_price = self.external_price_feed[tkn]['price_usd'][-1]
        bnt_rate = tkn_price / bnt_price
        return bnt_rate

    def set_whitelisted_tokens(self, whitelisted_tokens, bootstrap_tokens=True):
        """Check if the spot rate is outside of the EMA tolerance. No changes are made to the pool depth if outside limits.

        Args:
            whitelisted_tokens (list): The tickername of the token.
            bootstrap_tokens (bool): If true, automatically initializes all whitelisted tokens and adds them to the bootstrapped_tokens list.
        Returns:
            bnTKN_value (int or float or decimal): value of a single bnTKN pool token.
        """
        self.whitelisted_tokens = whitelisted_tokens
        if bootstrap_tokens:
            self.init_pool(self.whitelisted_tokens)

    def init_pool(self, tkns):
        """Initializes a new pool in the respective ledgers but does not perform the DAOmsig initialization which allows trading.

        Args:
            tkns (list): List of tickernames of the tokens to be initialized.
        Returns:
            self
        """
        for tkn in tkns:
            self.vault_ledger[tkn] = [0]
            self.pool_token_supply_ledger[f"bn{tkn}_ERC20_contract"] = {'block_num': [0], 'supply': [0]}
            self.staking_ledger[tkn] = [0]
            self.external_protection_wallet_ledger[tkn] = [0]

    def dao_msig_initialize_pools(self):
        """Command to initialize the pools in simulation of DAOmsig actions.

        Returns:
            self
        """
        dao_msig_initialized_pools = self.dao_msig_initialized_pools
        for TKN in self.whitelisted_tokens:
            BNTperTKN = self.get_bnt_rate(TKN)
            if self.available_liquidity_ledger[TKN][TKN][-1] == 0 and self.available_liquidity_ledger[TKN]['BNT'][-1] == 0 and self.vault_ledger[TKN][
                -1] * BNTperTKN > self.min_liquidity_threshold and (TKN not in self.dao_msig_initialized_pools):
                self.available_liquidity_ledger[TKN]['block_num'].append(self.available_liquidity_ledger[TKN]['block_num'][-1])
                self.available_liquidity_ledger[TKN][TKN].append(self.min_liquidity_threshold / BNTperTKN)
                self.available_liquidity_ledger[TKN]['BNT'].append(self.min_liquidity_threshold)
                self.vault_ledger['BNT'][-1] += self.min_liquidity_threshold
                self.staking_ledger['BNT'][-1] += self.min_liquidity_threshold
                self.pool_token_supply_ledger[f"bnBNT_ERC20_contract"]["supply"][-1] += self.min_liquidity_threshold
                self.update_ema(TKN)
                self.available_liquidity_ledger[TKN]['funding_limit'].append(self.available_liquidity_ledger[TKN]['funding_limit'][-1] - self.min_liquidity_threshold)
                dao_msig_initialized_pools.append(TKN)
            else:
                pass
        self.dao_msig_initialized_pools = dao_msig_initialized_pools
        self.check_bootstrapped_tokens()

    def update_ema(self, tkn):
        """Updates the EMA (Exponential Moving Average) based on current spot rates.

        Args:
            tkn (str): The tickername of the token.
        Returns:
            self
        """
        ema = self.ema
        if tkn in self.ema:
            if ema[tkn]['block_num'][-1] < self.block_num:
                ema[tkn]['block_num'].append(self.block_num)
                ema[tkn]['ema'].append(self.alpha * self.get_spot_rate(tkn) + self.ema[tkn]['ema'][-1] * (1 - self.alpha))
        else:
            ema[tkn]['block_num'].append(self.block_num)
            ema[tkn]['ema'].append(self.exchange_rates[tkn])
        self.ema = ema

    def is_within_ema_tolerance(self, tkn, lower_limit=.99, upper_limit=1.01):
        """Check if the spot rate is outside of the EMA tolerance. No changes are made to the pool depth if outside limits.

        Args:
            tkn (str): The tickername of the token.
        Returns:
            bnTKN_value (int or float or decimal): value of a single bnTKN pool token.
        """
        spot_rate = self.get_spot_rate(tkn)
        ema = self.ema[tkn]['ema'][-1]
        result = (lower_limit * ema) <= spot_rate <= (upper_limit * ema)
        if self.verbose:
            print(
                f"spot_rate={spot_rate} \n",
                f"ema={ema} \n",
                f"EMA Test = {result}, ({lower_limit} * {ema}) <= {spot_rate} <= ({upper_limit} * {ema}), tkn={tkn}"
            )
        return result

    def add_tkn_to_vault(self, tkn, x):
        """Adds x tkn to the self.vault_ledger dictionary with the same block_num number.

        Args:
            tkn (str): The tickername of the token.
            x (int or float or decimal): The amount of tkn.
        Returns:
            self
        """
        for key in self.vault_ledger:
            if key == tkn:
                self.vault_ledger[key].append(self.vault_ledger[key][-1] + x)
            else:
                self.vault_ledger[key].append(self.vault_ledger[key][-1])
            self.vault_ledger['block_num'][-1] = self.block_num

    def add_tkn_to_staking_ledger(self, tkn, x):
        """Adds x tkn to the self.staking_ledger dictionary with the same block_num number.

        Args:
            tkn (str): The tickername of the token.
            x (int or float or decimal): The amount of tkn.
        Returns:
            self
        """
        for key in self.staking_ledger:
            if key == tkn:
                self.staking_ledger[key].append(self.staking_ledger[key][-1] + x)
            else:
                self.staking_ledger[key].append(self.staking_ledger[key][-1])
            self.staking_ledger['block_num'][-1] = self.block_num

    def add_tkn_to_vortex_ledger(self, x):
        """Adds TKN to the vortex ledger.

        Args:
            x (int or float or decimal): The value of BNT to add to the vortex.

        Returns:
            self
        """
        self.vortex_ledger['BNT'].append(self.vortex_ledger['BNT'][-1] + x)
        self.vortex_ledger['block_num'].append(self.block_num)

    def add_tkn_to_pool_token_supply_ledger(self, tkn, x):
        self.pool_token_supply_ledger[tkn]['block_num'].append(self.block_num)
        for key in self.pool_token_supply_ledger:
            if key == tkn:
                self.pool_token_supply_ledger[tkn]['supply'].append(self.pool_token_supply_ledger[tkn]['supply'][-1] + x)
            else:
                self.pool_token_supply_ledger[key]['supply'].append(self.pool_token_supply_ledger[key]['supply'][-1])

    def add_tkn_to_available_liquidity_ledger(self, tkn, x):
        """Adds TKN to the available liquidity ledger.

        Args:
            tkn (str): The tickername of the token.
            x (int or float or decimal): The value of liquidity to add.

        Returns:
            self
        """
        for key in self.available_liquidity_ledger:
            if key == tkn:

                a = self.available_liquidity_ledger[key]['BNT'][-1]
                b = self.available_liquidity_ledger[key][tkn][-1]
                c = self.vault_ledger[key][-1] - b
                spot_rate = self.get_spot_rate(tkn)
                funding_limit = self.available_liquidity_ledger[key]['funding_limit'][-1]

                if (c >= b) & (a <= funding_limit):
                    self.available_liquidity_ledger[key]['BNT'].append(a * 2)
                    self.available_liquidity_ledger[key][tkn].append(b * 2)
                    self.available_liquidity_ledger[key]['funding_limit'].append(funding_limit - a)

                elif (c >= b) & (a > funding_limit):
                    self.available_liquidity_ledger[key]['BNT'].append(a + funding_limit)
                    self.available_liquidity_ledger[key][tkn].append(funding_limit / spot_rate)
                    self.available_liquidity_ledger[key]['funding_limit'].append(0)

                elif (c < b) & (a <= funding_limit):
                    self.available_liquidity_ledger[key]['BNT'].append((b + c) * spot_rate)
                    self.available_liquidity_ledger[key][tkn].append(b + c)
                    self.available_liquidity_ledger[key]['funding_limit'].append(funding_limit - (c * spot_rate))

                else:
                    self.available_liquidity_ledger[key]['BNT'].append(a + (c * spot_rate))
                    self.available_liquidity_ledger[key][tkn].append(b + c)
                    self.available_liquidity_ledger[key]['funding_limit'].append(funding_limit - (c * spot_rate))

                self.available_liquidity_ledger[key]['block_num'].append(self.block_num)
                final_liquidity = self.available_liquidity_ledger[key]['BNT'][-1]
                delta = final_liquidity - a
                self.add_tkn_to_vault('BNT', delta)
                self.add_tkn_to_staking_ledger('BNT', delta)
                bnt_bnbnt_rate = self.staking_ledger['BNT'][-1] / self.pool_token_supply_ledger[f"bnBNT_ERC20_contract"]["supply"][-1]
                update_amt = delta / bnt_bnbnt_rate
                self.add_tkn_to_pool_token_supply_ledger("bnBNT_ERC20_contract", update_amt)

                if "bnBNT" not in self.protocol_agent.wallet:
                    self.protocol_agent.wallet["bnBNT"] = {'block_num':[0],
                                                           "bnBNT": [0]}

                self.protocol_agent.wallet["bnBNT"]["bnBNT"].append(self.protocol_agent.wallet["bnBNT"]["bnBNT"][-1] + update_amt)
                self.protocol_agent.wallet["bnBNT"]['block_num'].append(self.block_num)

            else:
                self.available_liquidity_ledger[key][key].append(self.available_liquidity_ledger[key][key][-1])
                self.available_liquidity_ledger[key]['BNT'].append(self.available_liquidity_ledger[key]['BNT'][-1])
                self.available_liquidity_ledger[key]['block_num'].append(self.block_num)
                self.available_liquidity_ledger[key]['funding_limit'].append(self.available_liquidity_ledger[key]['funding_limit'][-1])

    def mint_bntkn(self, tkn, x=None, bnt=None):
        """Adds TKN to the available liquidity ledger.

        Args:
            tkn (str): The tickername of the token.
            x (int or float or decimal): The value of liquidity to add. (default=None)
            bnt (int or float or decimal): The value of bnt liquidity to add. (default=None)

        Returns:
            y (int): The number of pool tokens issued to the user.
        """
        if bnt is None: y = float(x) * float(self.get_pool_token_value(tkn))
        else: y = float(bnt)
        for contract in self.pool_token_supply_ledger:
            self.pool_token_supply_ledger[contract]['block_num'].append(self.block_num)
            if contract == f'bn{tkn}_ERC20_contract':
                self.pool_token_supply_ledger[contract]['supply'].append(self.pool_token_supply_ledger[contract]['supply'][-1] + y)
            else:
                self.pool_token_supply_ledger[contract]['supply'].append(self.pool_token_supply_ledger[contract]['supply'][-1])
        return (y)

    def check_pool_shutdown(self, warning_only=True):
        """Checks if any whitelisted tokens meet the pool-shutdown criterea, and resets the DOAmsig state if yes.

        Returns:
            self
        """
        dao_msig_initialized_pools = self.dao_msig_initialized_pools
        for TKN in self.whitelisted_tokens:
            BNTperTKN = self.get_bnt_rate(TKN)
            try:
                if self.vault_ledger[TKN][-1] < 0 or self.available_liquidity_ledger[TKN][-1] < 0 or self.staking_ledger[TKN][-1] < 0 or self.external_protection_wallet_ledger[TKN][-1] < 0:
                    if not warning_only:
                        raise UserWarning(f'Error. {TKN} balance cannot be less than zero. {self.vault_ledger[TKN][-1]}')
                    #     dao_msig_initialized_pools = [token for token in dao_msig_initialized_pools if token != TKN]
                    #     self.available_liquidity_ledger[TKN]['BNT'].append(0)
                    #     self.available_liquidity_ledger[TKN][TKN].append(0)
                    #     self.available_liquidity_ledger[TKN]['funding_limit'].append(self.bnt_funding_limit)
                    #     self.available_liquidity_ledger[TKN]['block_num'].append(
                    #         self.available_liquidity_ledger[TKN]['block_num'][-1])
                    #     self.dao_msig_initialized_pools = dao_msig_initialized_pools
            except:
                pass

            if float(self.vault_ledger[TKN][-1]) * float(BNTperTKN) < self.min_liquidity_threshold and (
                        TKN in self.dao_msig_initialized_pools):
                if self.verbose:
                    print(
                        f"WARNING POOL SHUTDOWN!!! \n",
                        f"warning_only={warning_only} \n",
                        f"pool={TKN} \n",
                        f"TKN Liquidity <= Minimum Pool Liquidity, if not warning_only then will reset pool to pre-DAOmsig state"
                    )

    def check_bootstrapped_tokens(self):
        """Checks if new tokens have been whitelisted and whether or not those token meet the min_liquidity_threshold,
        ensures token are added to the bootstrapped_tokens list if any whitelisted tokens meet this criterea

        Returns:
            self
        """
        bootstrapped_tokens = self.bootstrapped_tokens
        for tkn in self.whitelisted_tokens:
            if tkn != 'BNT':
                tkn_val = self.vault_ledger[tkn][-1] * self.exchange_rates[tkn]
                if (tkn_val >= self.min_liquidity_threshold) & (tkn not in bootstrapped_tokens) & (tkn in self.dao_msig_initialized_pools):
                    bootstrapped_tokens.append(tkn)
        self.bootstrapped_tokens = bootstrapped_tokens

    def check_case(self,
                   is_surplus,
                   is_deficit,
                   satisfies_hlim,
                   satisfies_hmax,
                   reduce_trading_liquidity,
                   a, b, c, e, w):
        """Defines which BIP15 logical case applies to the current action.

        Args:
            is_surplus (bool): Whether the system is in a surplus state.
            is_deficit (bool): Whether the system is in a deficit state.
            satisfies_hlim (bool): Whether the system passes the hlim test.
            satisfies_hmax (bool): Whether the system passes the hmax test.
            reduce_trading_liquidity (bool): Whether the system needs to reduce the available trading liquidity.
            a (int or float or decimal): the BNT balance of the trading liquidity, as judged from the spot rate.
            b (int or float or decimal): the TKN balance of the trading liquidity, as judged from the spot rate.
            c (int or float or decimal): the difference between the TKN balance of the vault, and the TKN trading liquidity.
            e (int or float or decimal): the TKN balance of the staking ledger.
            w (int or float or decimal): the TKN balance of the external protection wallet.

        Returns:
            case (str): the human readable logical-case name
        """
        if is_surplus & (a == 0 and b == 0):
            case = 'bootstrap surplus (special case)'

        elif is_deficit & (a == 0 and b == 0):
            case = 'bootstrap deficit (special case)'

        elif is_surplus & (satisfies_hlim and satisfies_hmax) & ((b + c) > e):
            case = 'arbitrage surplus'

        elif is_surplus & (not satisfies_hlim or not satisfies_hmax) & (not reduce_trading_liquidity):
            case = 'default surplus'

        elif is_surplus & (not satisfies_hlim or not satisfies_hmax) & reduce_trading_liquidity:
            case = 'bootstrap surplus'

        elif is_deficit & (satisfies_hlim and satisfies_hmax):
            case = 'arbitrage deficit'

        elif is_deficit & (not satisfies_hlim or not satisfies_hmax) & (
                not reduce_trading_liquidity):
            case = 'default deficit'

        elif is_deficit & (not satisfies_hlim or not satisfies_hmax) & reduce_trading_liquidity:
            case = 'bootstrap deficit'

        return case

    def validate_input_types(self, tkn, x):
        """Validates input data types using pydantic type setting. See utils/schema.py for details

        Args:
            tkn (str): The tickername of the token.
            x (int or float or decimal): The value of liquidity to add. (default=None)

        Returns:
            a (int or float or decimal): the BNT balance of the trading liquidity, as judged from the spot rate.
            b (int or float or decimal): the TKN balance of the trading liquidity, as judged from the spot rate.
            c (int or float or decimal): the difference between the TKN balance of the vault, and the TKN trading liquidity.
            e (int or float or decimal): the TKN balance of the staking ledger.
            w (int or float or decimal): the TKN balance of the external protection wallet.
            m (float or decimal): the global system exit fee
            n (float or decimal): the per pool trading fee
            x (int or float or decimal): the TKN or BNT amount involved in the current action
        """
        a = self.available_liquidity_ledger[tkn]['BNT'][-1]
        b = self.available_liquidity_ledger[tkn][tkn][-1]
        e = self.staking_ledger[tkn][-1]
        w = self.external_protection_wallet_ledger[tkn][-1]
        c = self.vault_ledger[tkn][-1] - b
        n = self.exit_fee
        m = self.pool_fees[tkn]
        data_validator = FixedPointUnstakeTKN if self.is_solidity else FloatingPointUnstakeTKN
        input_params = dict(
            a=a,
            b=b,
            e=e,
            w=w,
            c=c,
            n=n,
            m=m,
            x=x
        )
        self.validated_params = data_validator(**input_params)
        self.params_set = True
        return self.validated_params.a, self.validated_params.b, self.validated_params.c, self.validated_params.e, self.validated_params.m, self.validated_params.n, self.validated_params.x, self.validated_params.w

    def update_available_liquidity_ledger(self, tkn, bnt_trading_liquidity, tkn_trading_liquidity):
        """Updates available liquidity for swaps

        Args:
            tkn (str): The tickername of the token.
            bnt_trading_liquidity (int or float or decimal): The value of liquidity to add to bnt.
            tkn_trading_liquidity (int or float or decimal): The value of liquidity to add to tkn.

        Returns:
            self
        """
        # print('tkn_a=', tkn, 'bnt_trading_liquidity=', bnt_trading_liquidity)
        self.available_liquidity_ledger[tkn]['BNT'].append(bnt_trading_liquidity)
        self.available_liquidity_ledger[tkn][tkn].append(tkn_trading_liquidity)
        self.available_liquidity_ledger[tkn]['funding_limit'].append(
            self.available_liquidity_ledger[tkn]['funding_limit'][-1])
        self.available_liquidity_ledger[tkn]['block_num'].append(
            self.available_liquidity_ledger[tkn]['block_num'][-1])
        for key in self.available_liquidity_ledger:
            if key != tkn:
                self.available_liquidity_ledger[key]['funding_limit'].append(
                    self.available_liquidity_ledger[key]['funding_limit'][-1])
                self.available_liquidity_ledger[key]['block_num'].append(
                    self.available_liquidity_ledger[key]['block_num'][-1])
                self.available_liquidity_ledger[key]['BNT'].append(self.available_liquidity_ledger[key]['BNT'][-1])
                self.available_liquidity_ledger[key][key].append(self.available_liquidity_ledger[key][key][-1])

    def update_ledger_balances(self, tkn, p, q, r, s, t, u, a, b, c, e, m, n, x, w, case, event_type='unstake_tkn'):
        """Updates the ledger balances for withdraw actions

        Args:
            tkn (str): The tickername of the token.
            x (int or float or decimal): The value of liquidity to add. (default=None)
            a (int or float or decimal): the BNT balance of the trading liquidity, as judged from the spot rate.
            b (int or float or decimal): the TKN balance of the trading liquidity, as judged from the spot rate.
            c (int or float or decimal): the difference between the TKN balance of the vault, and the TKN trading liquidity.
            e (int or float or decimal): the TKN balance of the staking ledger.
            w (int or float or decimal): the TKN balance of the external protection wallet.

        Returns:
            self
        """
        if event_type == 'unstake_tkn':

            # handle surplus signs
            if (case != 'default surplus') & self.tokenomics.surplus:
                self.available_liquidity_ledger[tkn]['BNT'][-1] -= float(p)
                self.available_liquidity_ledger[tkn][tkn][-1] += float(r)
            elif self.tokenomics.surplus:
                self.available_liquidity_ledger[tkn]['BNT'][-1] -= float(p)
                self.available_liquidity_ledger[tkn][tkn][-1] -= float(r)

            # handle deficit signs
            elif (self.tokenomics.deficit) & ('ep2' in case):
                self.available_liquidity_ledger[tkn]['BNT'][-1] -= float(p)
                self.available_liquidity_ledger[tkn][tkn][-1] -= float(r)
            elif self.tokenomics.deficit:
                self.available_liquidity_ledger[tkn]['BNT'][-1] += float(p)
                self.available_liquidity_ledger[tkn][tkn][-1] -= float(r)

            self.staking_ledger['BNT'][-1] -= float(q)
            self.staking_ledger[tkn][-1] -= float(x)
            self.vault_ledger[tkn][-1] -= float(s)
            self.mint_bntkn(tkn, bnt=t)
            self.external_protection_wallet_ledger[tkn][-1] -= float(u)

        eq_case = self.tokenomics.reduce_trading_liquidity
        if self.tokenomics.surplus: eq = 'x(1-n) <= c'
        else: eq = 'x(1-n)(b+c)/e <= c'
        if self.tokenomics.satisfies_hlim and self.tokenomics.satisfies_hmax:
            eq_entry = ''
        else:
            eq_entry = f'{eq}:{eq_case}'
        self.describe_withdrawal = {
            'inputs': [
                f"a:{a}",
                f"b:{b}",
                f"c:{c}",
                f"e:{e}",
                f"m:{m}",
                f"n:{n}",
                f"x:{x}",
            ],
            'tests': [
                f'hlim:{self.tokenomics.hlim}',
                f'hmax:{self.tokenomics.hmax}',
                f'is_surplus:{self.tokenomics.surplus}',
                f'case={case}',
                f'satisfies_hlim:{self.tokenomics.satisfies_hlim}',
                f'satisfies_hmax:{self.tokenomics.satisfies_hmax}',
                eq_entry,
            ],
            'outputs': [
                f"p:{p}",
                f"q:{q}",
                f"r:{r}",
                f"s:{s}",
                f"t:{t}",
                f"u:{u}",
                '',
            ],
            'trading_liquidity': [
                f"BNT:{self.available_liquidity_ledger[tkn]['BNT'][-1]}",
                f"{tkn}:{self.available_liquidity_ledger[tkn][tkn][-1]}",
                '',
                '',
                '',
                '',
                '',
            ],
            'vault_ledger': [
                f"{tkn}:{self.vault_ledger[tkn][-1]}",
                '',
                '',
                '',
                '',
                '',
                '',
            ],
            'staking_ledger': [
                f"{tkn}:{self.staking_ledger[tkn][-1]}",
                '',
                '',
                '',
                '',
                '',
                '',
            ],
            'external_protection_ledger': [
                f"{tkn}:{self.external_protection_wallet_ledger[tkn][-1]}",
                '',
                '',
                '',
                '',
                '',
                '',
            ]
        }
        return self

    def describe(self, withdraw=False):
        """Outputs details of the system state, including ledgers & parameters, or the case of withdrawal, inputs, outputs, &etc.
        Args:
            return_dict (bool): If True, return dictionary of information about system state. Defaults to False.
        Returns:
            pd.DataFrame: DataFrame of system state
        """
        if withdraw:
            logger = self.describe_withdrawal
        else:
            # try:
            logger = {'trading_liquidity':[],
                      'vault_ledger': [],
                      'staking_ledger': [],
                      'pool_token_supply_ledger': [],
                      'vortex_ledger': []
                      }
            for tkn in self.available_liquidity_ledger:
                if tkn != 'BNT':
                    logger['trading_liquidity'].append(f"{tkn}:{self.available_liquidity_ledger[tkn][tkn][-1]} - BNT:{self.available_liquidity_ledger[tkn]['BNT'][-1]}")
            for tkn in self.available_liquidity_ledger:
                logger['vault_ledger'].append(
                    f"{tkn}:{self.vault_ledger[tkn][-1]}")
            for tkn in self.staking_ledger:
                if tkn != 'block_num':
                    logger['staking_ledger'].append(
                    f"{tkn}:{self.staking_ledger[tkn][-1]}")
            for tkn in self.pool_token_supply_ledger:
                logger['pool_token_supply_ledger'].append(
                    f"{tkn.replace('_ERC20_contract','')}:{self.pool_token_supply_ledger[tkn]['supply'][-1]}")

            logger['vortex_ledger'].append(f"Balance:{self.vortex_ledger['BNT'][-1]}")

            max_rows = max([len(logger['pool_token_supply_ledger']),
                            len(logger['staking_ledger']),
                            len(logger['vault_ledger']),
                            len(logger['trading_liquidity'])])

            for key in logger:
                for row in range(max_rows - len(logger[key])):
                    logger[key].append('')

            for ledger in logger:
                logger[ledger].sort()

        return pd.DataFrame(logger)




"""Tools for use in DeFi
"""
import matplotlib.pyplot as plt
import pandas as pd
import numpy as np
import datetime, requests
from matplotlib.gridspec import GridSpec
from pandas.plotting import register_matplotlib_converters

register_matplotlib_converters()


def iloss(price_ratio, numerical=False):
    """return the impermanent loss result in compare with buy&hold A&B assets


    Args:
        price_ratio (float): Variation A Asset / Variation B Asset

            price_ratio formula:
                price_ratio = (var_A/100 + 1) / (var_B/100 + 1)
                    var_A: Asset A % variation
                    var_B: Asset B % variation

        numerical (bool): if True, returns impermanent loss as a decimal expr, ie "5%"" => 0.05 (Default: False)

    Returns:
        TYPE: impermanent loss as a string percentual value
    """
    il = 2 * (price_ratio ** 0.5 / (1 + price_ratio)) - 1
    r = f"{il:.2%}" if not numerical else il

    return r


def compare(days, var_A=0, var_B=0, rw_pool_A=0, rw_pool_B=0, rw_pool_AB=0, fees_AB=0):
    """Compare for 2 assets, buy&hold strategy with separate staking and farming by liquidity pool providing.
        Considering: impermanent loss, fees earned and farming/staking rewards

    Args:
        days (int): days for strategy
        var_A (float, optional): Percentual variation for A token. Ex 10 for 10%
        var_B (float, optional): Percentual variation for B token. Ex 10 for 10%
        rw_pool_A (float, optional): Percentual rewards per day for one asset pool (Token A)
        rw_pool_B (float, optional): Percentual rewards per day for one asset pool (Token B)
        rw_pool_AB (float, optional): Percentual rewards per day for two asset farm (LP Token AB)
        fees_AB (float, optional): Percentual provider liquidity fees earned per day

    Returns:
        dict: Percentual returns for each strategy:
            buy_hold two assets in your wallet
            stake two assets at individual pools
            farming by liquidity pool
    """
    buy_hold = (0.5 * var_A + 0.5 * var_B) / 100
    x = (var_A / 100 + 1) / (var_B / 100 + 1)
    perdida_impermanente = 2 * (x ** 0.5 / (1 + x)) - 1

    stake = buy_hold + 0.5 * days * (rw_pool_A / 100 + rw_pool_B / 100)
    farm = buy_hold * (1 + perdida_impermanente) + days * (rw_pool_AB / 100 + fees_AB / 100)
    mejor = 'Farm' if farm > stake else 'Stake'

    return {'buy_hold': f'{buy_hold:.2%}', 'stake': f'{stake:.2%}', 'farm': f'{farm:.2%}', 'Best': mejor}


######################################################################
##                                                                  ##
##      Llama API                                                   ##
##      Public API https://docs.llama.fi/api                        ##
##                                                                  ##
######################################################################


def getProtocols():
    """Get list all DeFi components2 across all blockchains

    Returns:
        DataFrame: All DeFi dApps
    """
    url = "https://api.llama.fi/protocols"
    r = requests.get(url)
    r_json = r.json()
    df = pd.DataFrame(r_json)
    df.set_index('name', inplace=True)
    return df


def getProtocol(protocol):
    """Get metrics and historic TVL for one DeFi dApp

    Args:
        protocol (String): Name of protocol ie "Uniswap"

    Returns:
        tuple (Dictionary, DataFrame): Dictionary with protocol metadata & DataFrame with historical TVL
    """
    url = f"https://api.llama.fi/protocol/{protocol}"
    r = requests.get(url)
    r_json = r.json()

    df = pd.DataFrame(r_json['tvl'])
    df.date = pd.to_datetime(df.date, unit='s')
    df = df.set_index('date')
    del r_json['tvl']
    metadata = r_json

    return metadata, df


def getChart():
    """Get historical TVL across all DeFi dApps, cummulative result

    Returns:
        DataFrame: DataFrame date-indexed with all days TVL
    """

    url = "https://api.llama.fi/charts"
    r = requests.get(url)
    r_json = r.json()
    df = pd.DataFrame(r_json)
    df.date = pd.to_datetime(df.date, unit='s')
    df = df.set_index('date')
    return df


######################################################################
##                                                                  ##
##      CoinGecko API                                               ##
##      Public API https://www.coingecko.com/es/api                 ##
##                                                                  ##
######################################################################

def gecko_list(page=1, per_page=250):
    """Returns list of full detail conGecko currency list

    Args:
        page (int, optional): number of pages
        per_page (int, optional): number of records per page

    Returns:
        DataFrame: list of full detail conGecko currency list
    """
    url = "https://api.coingecko.com/api/v3/coins/markets"
    params = {"vs_currency": "usd", "order": "market_cap_desc", "per_page": per_page, "page": page}
    r = requests.get(url, params).json()
    df = pd.DataFrame(r)
    df.set_index('symbol', inplace=True)
    return df


def get_gecko_spot_price(tokens, quote):
    """get price of combine pairs

    Args:
        tokens (comma separated strings): ie "bitcoin,ethereum"
        quote (comma separated fiat or quote currency): ie: "usd,eur"

    Returns:
        dictionary: Returns pairs quotes
    """

    url = "https://api.coingecko.com/api/v3/simple/price"
    params = {"ids": tokens, "vs_currencies": quote}
    r = requests.get(url, params).json()
    return r


def geckoMarkets(ticker):
    """Get top100 markets (pairs, quotes, exchanges, volume, spreads and more)

    Args:
        ticker (string): gecko ID, ie "bitcoin"

    Returns:
        DataFrame: Full detail markets available
    """
    url = f"https://api.coingecko.com/api/v3/coins/{ticker}/tickers"
    r = requests.get(url).json()['tickers']
    df = pd.DataFrame(r)
    df['exchange'] = df['market'].apply(pd.Series)['name']
    df['volume_usd'] = df['converted_volume'].apply(pd.Series)['usd']
    df['price_usd'] = df['converted_last'].apply(pd.Series)['usd']

    df.set_index('exchange', inplace=True)
    cols = ['base', 'target', 'last', 'volume', 'bid_ask_spread_percentage', 'timestamp',
            'volume_usd', 'price_usd', 'trust_score']
    df = df.loc[:, cols]
    cols[4] = 'spread'
    df.columns = cols
    df.timestamp = pd.to_datetime(df.timestamp)

    return df.sort_values('volume_usd', ascending=False)


def geckoHistorical(ticker, vs_currency='usd', days='max'):
    """Historical prices from coinGecko

    Args:
        ticker (string): gecko ID, ie "bitcoin"
        vs_currency (str, optional): ie "usd" (default)
        days (str, optional): ie "20", "max" (default)

    Returns:
        DataFrame: Full history: date, price, market cap & volume
    """

    url = f"https://api.coingecko.com/api/v3/coins/{ticker}/market_chart"
    params = {"vs_currency": {vs_currency}, "days": days}
    r = requests.get(url, params).json()
    prices = pd.DataFrame(r['prices'])
    market_caps = pd.DataFrame(r['market_caps'])
    total_volumes = pd.DataFrame(r['total_volumes'])
    df = pd.concat([prices, market_caps[1], total_volumes[1]], axis=1)
    df[0] = pd.to_datetime(df[0], unit='ms')
    df.columns = ['date', 'price', 'market_caps', 'total_volumes']
    df.set_index('date', inplace=True)

    return df


def getGeckoIDs():
    """IDs List from coinGecko

    Returns:
        list: First 5000 coingecko IDs by marketCap rank
    """
    ids_list = []

    for i in range(20):
        print(f'searching coins page: {i}   ', end='\r')
        ids_list = ids_list + gecko_list(page=i, per_page=250)['id'].tolist()
    return ids_list


def farmSimulate(pair, apr, start='2021-01-01'):
    """Simulate farm result with historical prices & APR value

    Args:
        pair (list): gecko IDs list ["bitcoin",'tether']
        apr (float): ie 25 (for 25% Anual rewards)
        start (str, optional): ISO Format YYYY-MM-DD ie "2021-01-01", "2021-01-01" (default)

    Returns:
        Dict & Plot: Full farming strategy results
    """

    prices = pd.DataFrame()
    for coin in pair:
        print(f'Downloading {coin}')
        try:
            df = geckoHistorical(coin)
            prices[coin] = df['price']
        except:
            print(f'Error geting {coin} prices')

    if len(prices.columns) == 2:
        prices = prices.dropna().iloc[:]
        start = datetime.datetime.strptime(start, '%Y-%m-%d')
        farm = prices.loc[prices.index >= start]
        farm = farm.divide(farm.iloc[0])
        farm['ratio'] = farm.iloc[:, 1].divide(farm.iloc[:, 0])

        farm['iloss'] = 2 * (farm['ratio'] ** 0.5 / (1 + farm['ratio'])) - 1
        farm['rewards'] = pd.Series(2 * apr / 100 / 365, index=farm.index).cumsum()
        farm['buy_hold'] = (farm.iloc[:, 0] + farm.iloc[:, 1]) / 2
        farm['farm'] = farm.buy_hold - farm.iloss + farm.rewards

        cagrs = farm.iloc[-1] ** (1 / (365 / len(farm))) - 1
        sigmas = farm.pct_change().std() * 365 ** 0.5
        sharpes = cagrs.divide(sigmas).round(2)
        dd = farm / farm.cummax() - 1

        fig = plt.figure(figsize=(15, 8))
        gs = GridSpec(nrows=2, ncols=4, figure=fig, height_ratios=[2, 1], hspace=0.45, wspace=0.35, top=.9)
        ax_upleft = fig.add_subplot(gs[0, 0:2])
        ax_upright = fig.add_subplot(gs[0, 2:])
        cols = 4
        ax_down = [fig.add_subplot(gs[1, i]) for i in range(cols)]

        ax_upleft.plot(farm.iloss.abs(), label='Impermanent Loss')
        ax_upleft.plot(farm.rewards, label='Farming Rewards')
        ax_upleft.legend()
        ax_upleft.grid()
        ax_upleft.set_title('Impermanent Loss vs Farming Rewards')
        ax_upleft.tick_params(axis='x', rotation=45)

        ax_upright.plot(farm.iloc[:, :2])
        ax_upright.plot(farm.buy_hold)
        ax_upright.plot(farm.farm)
        ax_upright.grid()
        ax_upright.legend([pair[0], pair[1], 'Buy&Hold', 'Farming Strategy'])
        ax_upright.set_title(f'{pair[0]} vs {pair[1]} vs Buy & Hold vs Farming strategy payoff')
        ax_upright.tick_params(axis='x', rotation=45)

        cagrs[[pair[0], pair[1], 'buy_hold', 'farm']].plot(kind='bar', ax=ax_down[0])
        sigmas[[pair[0], pair[1], 'buy_hold', 'farm']].plot(kind='bar', ax=ax_down[1])
        sharpes[[pair[0], pair[1], 'buy_hold', 'farm']].plot(kind='bar', ax=ax_down[2])
        dd[[pair[0], pair[1], 'buy_hold', 'farm']].min().plot(kind='bar', ax=ax_down[3])

        ax_down[0].set_title('CAGR', fontsize=12)
        ax_down[1].set_title('Anualized Volatility', fontsize=12)
        ax_down[2].set_title('Sharpe Ratio', fontsize=12)
        ax_down[3].set_title('Max DrawDowns', fontsize=12)
        [ax_down[i].grid(alpha=0.4) for i in range(cols)]

        for i in range(cols):
            ax_down[i].spines['top'].set_visible(False)
            ax_down[i].spines['right'].set_visible(False)

        b_h = (farm.iloc[-1].iloc[0] + farm.iloc[-1].iloc[1]) / 2 - 1
        iloss = farm.iloc[-1].iloss
        rewards = farm.iloc[-1].rewards
        net_farming = b_h - iloss + rewards

        result = {'Token 1': pair[0], 'Token 2': pair[1], 'start': start.isoformat()[:10],
                  'fixed APR': f'{apr / 100:.0%}', 'Buy & Hold': f'{b_h:.2%}',
                  'Impermanent Loss': f'{iloss:.2%}', 'Farming Rewards': f'{rewards:.2%}',
                  'Farming + Rewards - IL': f'{net_farming:.2%}'}

    else:
        result = 'Error geting historical prices, see geckoIDs() function to get CoinGecko IDs'
    return result


######################################################################
##                                                                  ##
##       Pancake Swap API                                           ##
##       API https://github.com/pancakeswap/pancake-info-api        ##
##                                                                  ##
######################################################################


def toFloatPartial(df):
    for i in df.columns:
        try:
            df[[i]] = df[[i]].astype(float)
        except:
            pass
    return df


def pcsSummary(as_df=True):
    url = "https://api.pancakeswap.info/api/v2/summary"
    r = requests.get(url).json()
    data = r.get('data', None)
    upd = r.get('updated_at') / 1000
    upd_dt = datetime.datetime.fromtimestamp(upd)

    if as_df:
        df = pd.DataFrame.from_dict(data, orient='index')
        df = toFloatPartial(df)
        df['updated'] = upd_dt
        return df
    else:
        return r


def pcsTokens(as_df=True):
    """get all token listed in pancakeswap

    Args:
        as_df (bool, optional): if True (default), return is a dataframe, else is a dictionary

    Returns:
        DataFrame with next columns: name  symbol  price   price_BNB   updated

    """
    # ultimo precio y volumen de base/quote de todos los pares

    url = "https://api.pancakeswap.info/api/v2/tokens"
    r = requests.get(url).json()
    data = r.get('data', None)
    upd = r.get('updated_at') / 1000
    upd_dt = datetime.datetime.fromtimestamp(upd)

    if as_df:
        df = pd.DataFrame.from_dict(data, orient='index')
        df = toFloatPartial(df)
        df['updated'] = upd_dt
        return df
    else:
        return r


def pcsPairs(as_df=True):
    """get top 1000 pancakeswap pairs LP order by reserves

    Args:
        as_df (bool, optional): if True (default), return is a dataframe, else is a dictionary

    Returns:
        DataFrame with next columns: 'pair_address', 'base_name', 'base_symbol', 'base_address',
       'quote_name', 'quote_symbol', 'quote_address', 'price', 'base_volume',
       'quote_volume', 'liquidity', 'liquidity_BNB', 'updated'
    """

    url = "https://api.pancakeswap.info/api/v2/pairs"
    r = requests.get(url).json()
    data = r.get('data', None)
    upd = r.get('updated_at') / 1000
    upd_dt = datetime.datetime.fromtimestamp(upd)

    if as_df:
        df = pd.DataFrame.from_dict(data, orient='index')
        df = toFloatPartial(df)
        df['updated'] = upd_dt
        return df
    else:
        return r


def pcsTokenInfo(search):
    """get info from a token

    Args:
        search (string): Token symbol or contract address

    Returns:
        Dict:
        {
         'name': 'Wrapped BNB',
         'symbol': 'WBNB',
         'price': '524.5429',
         'price_BNB': '1'
         }
    """
    search = 'WBNB' if search.upper() == 'BNB' else search
    url = "https://api.pancakeswap.info/api/v2/tokens"
    r = requests.get(url).json()
    data = r.get('data', None)
    res = f"Not found: {search}"
    for contract, values in data.items():
        if search.upper() == values['symbol'].upper() or search.upper() == contract.upper():
            res = data[contract]
            break

    return res


def pcsPairInfo(base, quote):
    """get info from a token pair LP

    Args:
        base (string): Base LP token, ie "CAKE"
        quote (string): Quote LP token, ie "BNB"
        its the same if you call pcsPAirInfo('cake', 'bnb') or pcsPAirInfo('bnb', 'cake')
    Returns:
        Dict: {
                 'pair_address': '0xA527a61703D82139F8a06Bc30097cC9CAA2df5A6',
                 'base_name': 'PancakeSwap Token',
                 'base_symbol': 'Cake',
                 'base_address': '0x0E09FaBB73Bd3Ade0a17ECC321fD13a19e81cE82',
                 'quote_name': 'Wrapped BNB',
                 'quote_symbol': 'WBNB',
                 'quote_address': '0xbb4CdB9CBd36B01bD1cBaEBF2De08d9173bc095c',
                 'price': '0.04311198194009326668',
                 'base_volume': '22248744.85',
                 'quote_volume': '934856.36',
                 'liquidity': '982769040.63',
                 'liquidity_BNB': '1878155.84'
                }
                * price is actually a ratio between base/quote tokens
    """
    url = "https://api.pancakeswap.info/api/pairs"
    r = requests.get(url).json()
    data = r.get('data', None)
    res = f"Not found: {base}-{quote}"
    base = 'WBNB' if base.upper() == 'BNB' else base
    quote = 'WBNB' if quote.upper() == 'BNB' else quote

    for contract, values in data.items():
        base_ = base.upper() == values['base_symbol'].upper()
        quote_ = quote.upper() == values['quote_symbol'].upper()
        base_cross = base.upper() == values['quote_symbol'].upper()
        quote_cross = quote.upper() == values['base_symbol'].upper()

        if (base_ and quote_) or (base_cross and quote_cross):
            res = data[contract]
            break

    return res


# from scipy import interpolate
import matplotlib.cm as cm


# def iloss_simulate(base_token, quote_token, value=100, base_pct_chg=0, quote_pct_chg=0):
#     """Calculate simulated impermanent loss from an initial value invested, get real time prices from pancakeswap API
#         This method create a 3D interpolated surface for impermanent loss and initial/final value invested
#
#     Args:
#         base_token (string): Pair first token, ie CAKE
#         quote_token (string): Pais second token, ie BNB
#         value (int, optional): Value investen in LP default=100
#         base_pct_chg (int, optional): value assming will change first token of LP pair, ie 10 (for +10% change)
#         quote_pct_chg (int, optional): value assming will change first token of LP pair, ie -30 (for -30% change)
#
#     Returns:
#         tuple (value_f, iloss): final value of value invested, and decimal impermanent loss
#     """
#     base_token = 'WBNB' if base_token.upper() == 'BNB' else base_token
#     quote_token = 'WBNB' if quote_token.upper() == 'BNB' else quote_token
#
#     # get real time prices
#     tokens = pcsTokens()
#     px_base = float(tokens.loc[tokens.symbol.str.upper() == base_token.upper()].price)
#     px_quote = float(tokens.loc[tokens.symbol.str.upper() == quote_token.upper()].price)
#
#     # Prepare grid
#     q_base, q_quote = (value / 2) / px_base, (value / 2) / px_quote
#     px_base, px_quote, q_base, q_quote
#     pxs_base = [px_base * i / 100 for i in range(1, 301)]
#     pxs_quote = [px_quote * i / 100 for i in range(1, 301)]
#     rows = []
#     for px_b in pxs_base:
#         for px_q in pxs_quote:
#             ratio = (px_b / px_base) / (px_q / px_quote)
#             iloss = 2 * (ratio ** 0.5 / (1 + ratio)) - 1
#
#             row = {'px_base': px_b, 'px_quote': px_q,
#                    'ratio': (px_b / px_base) / (px_q / px_quote),
#                    'impremante_loss': iloss}
#             rows.append(row)
#     df = pd.DataFrame(rows)
#     df_ok = df.loc[:, ['px_base', 'px_quote', 'impremante_loss']]
#     df_ok = df_ok.replace('NaN', np.nan).dropna()
#
#     if all(isinstance(i, (int, float)) for i in (value, base_pct_chg, quote_pct_chg)):
#         px_base_f = px_base * (1 + base_pct_chg / 100)
#         px_quote_f = px_quote * (1 + quote_pct_chg / 100)
#         ratio = (px_base_f / px_base) / (px_quote_f / px_quote)
#         iloss = 2 * (ratio ** 0.5 / (1 + ratio)) - 1
#         value_f = (px_base_f * q_base + px_quote_f * q_quote) * (iloss + 1)
#     else:
#         px_base_f, px_quote_f = px_base, px_quote
#         iloss = 0
#         value_f = None
#         print('must input numerical amount and pct change for base and quote to calculations of final value')
#
#     # Ploting surface
#     fig = plt.figure(figsize=(8, 8))
#     x1 = np.linspace(df_ok['px_base'].min(), df_ok['px_base'].max(), len(df_ok['px_base'].unique()))
#     y1 = np.linspace(df_ok['px_quote'].min(), df_ok['px_quote'].max(), len(df_ok['px_quote'].unique()))
#     x2, y2 = np.meshgrid(x1, y1)
#     Z = interpolate.griddata((df_ok['px_base'], df_ok['px_quote']), df_ok['impremante_loss'], (x2, y2))
#     Z[np.isnan(Z)] = df_ok.impremante_loss.mean()
#     ax = plt.axes(projection='3d', alpha=0.2)
#     ax.plot_wireframe(x2, y2, Z, color='tab:blue', lw=1, cmap='viridis', alpha=0.6)
#
#     # Start values ploting
#     xmax = df_ok.px_base.max()
#     ymax = df_ok.px_quote.max()
#     ax.plot([px_base, px_base], [0, px_quote], [-1, -1], ls='--', c='k', lw=1)
#     ax.plot([px_base, px_base], [px_quote, px_quote], [0, -1], ls='--', c='k', lw=1)
#     ax.plot([px_base, 0], [px_quote, px_quote], [-1, -1], ls='--', c='k', lw=1)
#
#     # End values ploting
#     ax.plot([px_base_f, px_base_f], [0, px_quote_f], [-1, -1], ls='--', c='gray', lw=1)
#     ax.plot([px_base_f, px_base_f], [px_quote_f, px_quote_f], [iloss, -1], ls='--', c='gray', lw=1)
#     ax.plot([px_base_f, 0], [px_quote_f, px_quote_f], [-1, -1], ls='--', c='gray', lw=1)
#     ax.plot([px_base_f, px_base_f], [px_quote_f, ymax], [iloss, iloss], ls='--', c='gray', lw=1)
#     ax.plot([px_base_f, 0], [ymax, ymax], [iloss, iloss], ls='--', c='gray', lw=1)
#
#     # Plot settings
#     # Colorbar only for plot_surface() method instead plot_wireframe()
#     # m = cm.ScalarMappable(cmap=cm.viridis)
#     # m.set_array(df_ok['impremante_loss'])
#     # plt.colorbar(m, fraction=0.02, pad=0.1)
#     x, y, z = (px_base, px_quote, .05)
#     p = ax.scatter(x, y, z, c='k', marker='v', s=300)
#     ax.set_title('Impermanent Loss 3D Surface', y=0.95)
#     ax.set_xlabel(f'Price {base_token}')
#     ax.set_ylabel(f'Price {quote_token}')
#     ax.set_zlabel('Impremante loss')
#     ax.view_init(elev=25, azim=240)  # start view angle
#
#     print(f"\nStart value USD {value:.0f}, {base_token} USD {px_base:.2f}, {quote_token} USD {px_quote:.2f}")
#     print(f"\nResults assuming {base_token.upper()} {base_pct_chg}%, and {quote_token.upper()} {quote_pct_chg}%")
#     print(f"End value estimate USD {value_f:.0f}, iloss: {iloss:.2%}")
#     plt.show()
#
#     return value_f, iloss


from bancorml.environments.tokenomics import (
    VaultLedger,
    ExternalProtectionWalletLedger,
    PoolTokenSupplyLedger,
    StakingLedger,
    FixedPointTokenomics,
    FloatingPointTokenomics
)

from .bancor3 import Bancor3
from .env_base import EnvBase


"""Component that emulates a Vault Ledger."""
from .ledger_base import LedgerBase

class VaultLedger(LedgerBase):
    """Component that emulates a Vault Ledger."""
    def __init__(self):
        super().__init__()
        self._ledger = {'block_num': [0]}
